use super::*;

/// [`Deserializer`] with a cleaner interface for parsing multiple objects
/// from one stream ("inlining").
pub trait Inlining: Sized {
    /// Try to read `n` bytes. Consumes the deserializer on failure.
    fn iread_n<A, E>(
        self,
        n: usize,
        ok: impl FnOnce(&[u8]) -> A,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<(A, Self), E>;

    /// Read all bytes, consuming the deserializer.
    fn iread_all<A>(self, ok: impl FnOnce(&[u8]) -> A) -> A;

    /// Re-export of [`Deserializer::tell`].
    fn itell(&self) -> usize;
}

/// [`InliningExt::iread_n_const`].
pub trait InliningExt<E>: Inlining {
    /// `const`-length equivalent of [`Inlining::iread_n`].
    fn iread_n_const<const N: usize>(
        self,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<([u8; N], Self), E> {
        self.iread_n(
            N,
            |slice| slice.try_into().expect("`iread_n` wrong length"),
            err,
        )
    }
}

impl<D: Inlining, E> InliningExt<E> for D {}

impl<D: ?Sized + Deserializer> Inlining for &mut D {
    fn iread_n<A, E>(
        self,
        n: usize,
        ok: impl FnOnce(&[u8]) -> A,
        err: impl FnOnce(&[u8]) -> E,
    ) -> Result<(A, Self), E> {
        let slice = self.read_n(n);
        if slice.len() == n {
            Ok((ok(slice), self))
        } else {
            Err(err(slice))
        }
    }

    fn iread_all<A>(self, ok: impl FnOnce(&[u8]) -> A) -> A {
        ok(self.read_all())
    }

    fn itell(&self) -> usize {
        self.tell()
    }
}

/// [`InliningResultExt::seal`].
pub trait InliningResultExt<A, E> {
    /// Drop the deserializer from parsing result.
    fn seal(self) -> Result<A, E>;
}

impl<A, E, I: Inlining> InliningResultExt<A, E> for Result<(A, I), E> {
    fn seal(self) -> Result<A, E> {
        self.map(|(a, _)| a)
    }
}
